EXPORTS
  hsa_init
  hsa_shut_down
  hsa_system_get_info
  hsa_extension_get_name
  hsa_system_extension_supported
  hsa_system_major_extension_supported
  hsa_system_get_extension_table
  hsa_system_get_major_extension_table
  hsa_iterate_agents
  hsa_agent_get_info
  hsa_agent_get_exception_policies
  hsa_cache_get_info
  hsa_agent_iterate_caches
  hsa_agent_extension_supported
  hsa_agent_major_extension_supported
  hsa_queue_create
  hsa_soft_queue_create
  hsa_queue_destroy
  hsa_queue_inactivate
  hsa_queue_load_read_index_scacquire
  hsa_queue_load_read_index_relaxed
  hsa_queue_load_write_index_scacquire
  hsa_queue_load_write_index_relaxed
  hsa_queue_store_write_index_relaxed
  hsa_queue_store_write_index_screlease
  hsa_queue_cas_write_index_scacq_screl
  hsa_queue_cas_write_index_scacquire
  hsa_queue_cas_write_index_relaxed
  hsa_queue_cas_write_index_screlease
  hsa_queue_add_write_index_scacq_screl
  hsa_queue_add_write_index_scacquire
  hsa_queue_add_write_index_relaxed
  hsa_queue_add_write_index_screlease
  hsa_queue_store_read_index_relaxed
  hsa_queue_store_read_index_screlease
  hsa_agent_iterate_regions
  hsa_region_get_info
  hsa_memory_register
  hsa_memory_deregister
  hsa_memory_allocate
  hsa_memory_free
  hsa_memory_copy
  hsa_memory_assign_agent
  hsa_signal_create
  hsa_signal_destroy
  hsa_signal_load_relaxed
  hsa_signal_load_scacquire
  hsa_signal_store_relaxed
  hsa_signal_store_screlease
  hsa_signal_silent_store_relaxed
  hsa_signal_silent_store_screlease
  hsa_signal_wait_relaxed
  hsa_signal_wait_scacquire
  hsa_signal_group_create
  hsa_signal_group_destroy
  hsa_signal_group_wait_any_scacquire
  hsa_signal_group_wait_any_relaxed
  hsa_signal_and_relaxed
  hsa_signal_and_scacquire
  hsa_signal_and_screlease
  hsa_signal_and_scacq_screl
  hsa_signal_or_relaxed
  hsa_signal_or_scacquire
  hsa_signal_or_screlease
  hsa_signal_or_scacq_screl
  hsa_signal_xor_relaxed
  hsa_signal_xor_scacquire
  hsa_signal_xor_screlease
  hsa_signal_xor_scacq_screl
  hsa_signal_exchange_relaxed
  hsa_signal_exchange_scacquire
  hsa_signal_exchange_screlease
  hsa_signal_exchange_scacq_screl
  hsa_signal_add_relaxed
  hsa_signal_add_scacquire
  hsa_signal_add_screlease
  hsa_signal_add_scacq_screl
  hsa_signal_subtract_relaxed
  hsa_signal_subtract_scacquire
  hsa_signal_subtract_screlease
  hsa_signal_subtract_scacq_screl
  hsa_signal_cas_relaxed
  hsa_signal_cas_scacquire
  hsa_signal_cas_screlease
  hsa_signal_cas_scacq_screl
  hsa_isa_from_name
  hsa_agent_iterate_isas
  hsa_isa_get_info
  hsa_isa_get_info_alt
  hsa_isa_get_exception_policies
  hsa_isa_get_round_method
  hsa_wavefront_get_info
  hsa_isa_iterate_wavefronts
  hsa_isa_compatible
  hsa_code_object_serialize
  hsa_code_object_deserialize
  hsa_code_object_destroy
  hsa_code_object_get_info
  hsa_code_object_get_symbol
  hsa_code_object_get_symbol_from_name
  hsa_code_symbol_get_info
  hsa_code_object_iterate_symbols
  hsa_code_object_reader_create_from_file
  hsa_code_object_reader_create_from_memory
  hsa_code_object_reader_destroy
  hsa_executable_create
  hsa_executable_create_alt
  hsa_executable_destroy
  hsa_executable_load_code_object
  hsa_executable_load_program_code_object
  hsa_executable_load_agent_code_object
  hsa_executable_freeze
  hsa_executable_get_info
  hsa_executable_global_variable_define
  hsa_executable_agent_global_variable_define
  hsa_executable_readonly_variable_define
  hsa_executable_validate
  hsa_executable_validate_alt
  hsa_executable_get_symbol
  hsa_executable_get_symbol_by_name
  hsa_executable_symbol_get_info
  hsa_executable_iterate_symbols
  hsa_executable_iterate_agent_symbols
  hsa_executable_iterate_program_symbols
  hsa_status_string
  hsa_ext_program_create
  hsa_ext_program_destroy
  hsa_ext_program_add_module
  hsa_ext_program_iterate_modules
  hsa_ext_program_get_info
  hsa_ext_program_finalize
  hsa_amd_coherency_get_type
  hsa_amd_coherency_set_type
  hsa_amd_profiling_set_profiler_enabled
  hsa_amd_profiling_get_dispatch_time
  hsa_amd_profiling_async_copy_enable
  hsa_amd_profiling_get_async_copy_time
  hsa_amd_profiling_convert_tick_to_system_domain
  hsa_amd_signal_create
  hsa_amd_signal_wait_any
  hsa_amd_signal_async_handler
  hsa_amd_async_function
  hsa_amd_image_get_info_max_dim
  hsa_amd_queue_cu_set_mask
  hsa_amd_queue_cu_get_mask
  hsa_amd_memory_fill
  hsa_amd_memory_async_copy
  hsa_amd_memory_async_copy_on_engine
  hsa_amd_memory_copy_engine_status
  hsa_amd_memory_get_preferred_copy_engine
  hsa_amd_memory_async_copy_rect
  hsa_amd_memory_lock
  hsa_amd_memory_lock_to_pool
  hsa_amd_memory_unlock
  hsa_amd_agent_iterate_memory_pools
  hsa_amd_agent_memory_pool_get_info
  hsa_amd_agents_allow_access
  hsa_amd_memory_pool_get_info
  hsa_amd_memory_pool_allocate
  hsa_amd_memory_pool_free
  hsa_amd_memory_pool_can_migrate
  hsa_amd_memory_migrate
  hsa_amd_interop_map_buffer
  hsa_amd_interop_unmap_buffer
  hsa_amd_image_create
  hsa_ext_image_get_capability
  hsa_ext_image_data_get_info
  hsa_ext_image_create
  hsa_ext_image_import
  hsa_ext_image_export
  hsa_ext_image_copy
  hsa_ext_image_clear
  hsa_ext_image_destroy
  hsa_ext_sampler_create
  hsa_ext_sampler_create_v2
  hsa_ext_sampler_destroy
  hsa_ext_image_get_capability_with_layout
  hsa_ext_image_data_get_info_with_layout
  hsa_ext_image_create_with_layout
  hsa_amd_pointer_info
  hsa_amd_pointer_info_set_userdata
  hsa_amd_ipc_memory_create
  hsa_amd_ipc_memory_attach
  hsa_amd_ipc_memory_detach
  hsa_amd_ipc_signal_create
  hsa_amd_ipc_signal_attach
  hsa_amd_register_system_event_handler
  hsa_amd_queue_set_priority
  hsa_amd_register_deallocation_callback
  hsa_amd_deregister_deallocation_callback
  hsa_amd_signal_value_pointer
  _amdgpu_r_debug
  hsa_amd_svm_attributes_set
  hsa_amd_svm_attributes_get
  hsa_amd_svm_prefetch_async
  hsa_amd_spm_acquire
  hsa_amd_spm_release
  hsa_amd_spm_set_dest_buffer
  hsa_amd_portable_export_dmabuf
  hsa_amd_portable_close_dmabuf
  hsa_amd_vmem_address_reserve
  hsa_amd_vmem_address_reserve_align
  hsa_amd_vmem_address_free
  hsa_amd_vmem_handle_create
  hsa_amd_vmem_handle_release
  hsa_amd_vmem_map
  hsa_amd_vmem_unmap
  hsa_amd_vmem_set_access
  hsa_amd_vmem_get_access
  hsa_amd_vmem_export_shareable_handle
  hsa_amd_vmem_import_shareable_handle
  hsa_amd_vmem_retain_alloc_handle
  hsa_amd_vmem_get_alloc_properties_from_handle
  hsa_amd_agent_set_async_scratch_limit
  hsa_ven_amd_pcs_iterate_configuration
  hsa_ven_amd_pcs_create
  hsa_ven_amd_pcs_create_from_id
  hsa_ven_amd_pcs_destroy
  hsa_ven_amd_pcs_start
  hsa_ven_amd_pcs_stop
  hsa_ven_amd_pcs_flush
  hsa_amd_queue_get_info
  hsa_amd_enable_logging
  hsa_amd_signal_wait_all
  hsa_amd_portable_export_dmabuf_v2