/*
 * Copyright © Advanced Micro Devices, Inc., or its affiliates. 
 * 
 * SPDX-License-Identifier: MIT
 */
 
#include <iostream>
#include <algorithm>
#include <cmath>
#include <vector>
using namespace std;

double CalcMean(vector<double> scores);
double CalcMedian(vector<double> scores);
double CalcStdDeviation(vector<double> scores, int score_mean);
int CalcConcurrentQueues(vector<double> scores);



