/*
 * Copyright © Advanced Micro Devices, Inc., or its affiliates. 
 * 
 * SPDX-License-Identifier: MIT
 */
 
#include "base_test.hpp"

// Default Constructor
BaseTest::BaseTest(size_t num) {
  
  // Set the numIteration_ to be 10 by default
  num_iteration_ = num;
}

BaseTest::~BaseTest() {}

